
/*******************************************************************************
*
* File src_legs.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Inventory of source fields.
*
* The externally accessible functions are
*
*   void init_src_legs(src_legs_t *sl)
*     Initializes the structure sl. Every structure of type src_legs_t
*     should be initialized before it is passed as argument to other
*     programs.
*
*   void get_src_legs(theory_t th,int nd,ltree_t **ltd,src_legs_t *sl)
*     Finds the source legs of the diagrams ltd[0],..,ltd[nd-1] in the
*     theory th. The associated vertex indices and numbers of source legs
*     are assigned to the arrays in the structure sl (see the notes). The
*     arrays are automatically reallocated with sufficient size if needed.
*
* Notes:
*
* After calling get_src_legs(), the data in the structure sl provide a
* complete description of the sources attached to the vertices of the
* labeled tree diagrams ltd[0],..,ltd[nd-1]. For each type of source field,
* characterized by the field type and the source label, the program finds
* the vertices where they are attached and stores the information in the
* appropriate arrays.
*
* Since the vertices in the different diagrams must be distinguished, an
* index is assigned to them by counting the vertices in the first diagram,
* then those in the second, and so on. The first vertex in the first diagram
* thus has index 0 and the last vertex in the last diagram index n-1, where
* n is the total number of vertices.
*
* The elements of a structure of type src_legs_t are:
*
*   ntm      Maximal number nt of types of source fields that can be
*            accommodated in the structure.
*
*   nvm      Maximal total number of vertices of the tree diagrams,
*            which can be accommodated in the structure.
*
*   nt       Current number of types of source fields.
*
*   lt       Field types lt[i], i=0,..,nt-1, of the sources.
*
*   lbl      Labels lbl[i] of the sources.
*
*   nv       Numbers nv[i] of vertices, where the sources of type i
*            are attached.
*
*   ns       Total number ns[i] of sources of type i.
*
*   iv       Indices iv[i][j], j=0,..,nv[i]-1, of the vertices, where
*            the sources of type i are attached.
*
*   ne       Numbers ne[i][j], of the sources of type i attached at
*            the vertex with index iv[i][j].
*
* Fermion and anti-fermion sources are distinguished. The sources are ordered
* according to the field type SCALAR,GLUON,.. and secondly according to the
* source label. All sources, including the ones with label 0, are recorded by
* get_src_legs(). The root line (if present) is not counted as a source.
*
*******************************************************************************/

#define SRC_LEGS_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "rvt.h"
#include "contract.h"

static int nlbl[(int)(LTYPES)],ilbl[(int)(LTYPES)],olbl[(int)(LTYPES)];
static int clbl[(int)(LTYPES)];
static vlist_t vls;


void init_src_legs(src_legs_t *sl)
{
   (*sl).ntm=0;
   (*sl).nvm=0;
   (*sl).nt=0;

   (*sl).lbl=NULL;
   (*sl).nv=NULL;
   (*sl).ns=NULL;

   (*sl).iv=NULL;
   (*sl).ne=NULL;

   (*sl).lt=NULL;
}


static void free_src_legs(src_legs_t *sl)
{
   if ((*sl).ntm>0)
   {
      free((*sl).lt);
      free((*sl).iv);
      free((*sl).lbl);
   }
}


static void alloc_src_legs(int ntm,int nvm,src_legs_t *sl)
{
   int *lbl,**iv;
   ltype_t *lt;

   if ((ntm>(*sl).ntm)||(nvm>(*sl).nvm))
   {
      if ((*sl).ntm>ntm)
         ntm=(*sl).ntm;
      if ((*sl).nvm>nvm)
         nvm=(*sl).nvm;

      free_src_legs(sl);

      lbl=malloc((3*ntm+2*ntm*nvm)*sizeof(*lbl));
      iv=malloc(2*ntm*sizeof(*iv));
      lt=malloc(ntm*sizeof(*lt));

      error((lbl==NULL)||(iv==NULL)||(lt==NULL),1,
            "alloc_src_legs [src_legs.c]","Unable to allocate data arrays");

      (*sl).ntm=ntm;
      (*sl).nvm=nvm;
      (*sl).lbl=lbl;
      (*sl).iv=iv;
      (*sl).lt=lt;
   }
}


static void set_ilbl(theory_t th,int nd,ltree_t **ltd)
{
   int nv,nl,i,j,k;
   int (*v)[2],*nlv;
   ltype_t lt,lr,lt0;
   line_t *l;
   vertex_t *vx;
   diagram_t *dm;

   if (th==PHI4)
      lt0=SCALAR;
   else
      lt0=GLUON;

   for (i=0;i<(int)(LTYPES);i++)
      ilbl[i]=0;

   for (i=0;i<nd;i++)
   {
      dm=ltd[i][0].dm;
      nv=(*dm).nv;
      v=(*dm).v;

      for (j=0;j<nv;j++)
      {
         vx=vls.v[v[j][0]]+v[j][1];
         nlv=(*vx).nl;

         for (k=0;k<(int)(LTYPES);k++)
            ilbl[k]+=nlv[k];
      }

      nl=(*dm).nl;
      l=(*dm).l;

      for (j=0;j<nl;j++)
      {
         lt=l[j].type;
         lr=rvt[(int)(lt)];

         ilbl[(int)(lt)]-=1;
         ilbl[(int)(lr)]-=1;
      }

      nl=ltd[i][0].nl;
      l=ltd[i][0].l;

      for (j=0;j<nl;j++)
      {
         lt=l[j].type;
         lr=rvt[(int)(lt)];

         ilbl[(int)(lt)]-=1;
         ilbl[(int)(lr)]-=1;
      }

      if (ltd[i][0].rv<nv)
         ilbl[(int)(lt0)]-=1;
   }

   for (i=0;i<(int)(LTYPES);i++)
      ilbl[i]=(ilbl[i]>0);
}


static void set_nlbl(int nd,ltree_t **ltd)
{
   int nl,i,j;
   ltype_t lt,lr;
   line_t *l;

   for (i=0;i<(int)(LTYPES);i++)
      nlbl[i]=0;

   for (i=0;i<nd;i++)
   {
      for (j=0;j<(int)(LTYPES);j++)
         clbl[j]=0;

      nl=ltd[i][0].nl;
      l=ltd[i][0].l;

      for (j=0;j<nl;j++)
      {
         lt=l[j].type;
         lr=rvt[(int)(lt)];

         clbl[(int)(lt)]+=1;
         if (lr!=lt)
            clbl[(int)(lr)]+=1;
      }

      for (j=0;j<(int)(LTYPES);j++)
      {
         if (clbl[j]>nlbl[j])
            nlbl[j]=clbl[j];
      }
   }

   olbl[0]=0;

   for (i=1;i<(int)(LTYPES);i++)
      olbl[i]=olbl[i-1]+ilbl[i-1]+nlbl[i-1];
}


static void set_types(theory_t th,int nd,ltree_t **ltd,src_legs_t *sl)
{
   int nt,nv,i,j,k;
   diagram_t *dm;

   set_ilbl(th,nd,ltd);
   set_nlbl(nd,ltd);
   nt=0;

   for (i=0;i<(int)(LTYPES);i++)
      nt+=(ilbl[i]+nlbl[i]);

   nv=0;

   for (i=0;i<nd;i++)
   {
      dm=ltd[i][0].dm;
      nv+=(*dm).nv;
   }

   alloc_src_legs(nt,nv,sl);

   (*sl).nt=nt;
   (*sl).nv=(*sl).lbl+nt;
   (*sl).ns=(*sl).lbl+2*nt;

   (*sl).ne=(*sl).iv+nt;
   (*sl).iv[0]=(*sl).lbl+3*nt;
   (*sl).ne[0]=(*sl).lbl+3*nt+nt*nv;

   for (i=1;i<nt;i++)
   {
      (*sl).iv[i]=(*sl).iv[i-1]+nv;
      (*sl).ne[i]=(*sl).ne[i-1]+nv;
   }

   for (i=0;i<nt;i++)
   {
      (*sl).nv[i]=nv;

      for (j=0;j<nv;j++)
      {
         (*sl).iv[i][j]=j;
         (*sl).ne[i][j]=0;
      }
   }

   k=0;

   for (i=0;i<(int)(LTYPES);i++)
   {
      if (ilbl[i])
      {
         (*sl).lbl[k]=0;
         (*sl).lt[k]=(ltype_t)(i);
         k+=1;
      }

      for (j=1;j<=nlbl[i];j++)
      {
         (*sl).lbl[k]=j;
         (*sl).lt[k]=(ltype_t)(i);
         k+=1;
      }
   }
}


static void set_ne(theory_t th,int nd,ltree_t **ltd,src_legs_t *sl)
{
   int nv,nl,rv,i,j,k,m;
   int (*v)[2];
   ltype_t lt,lr,lt0;
   line_t *l;
   diagram_t *dm;
   vertex_t *vx;

   nv=0;

   for (i=0;i<nd;i++)
   {
      for (j=0;j<(int)(LTYPES);j++)
         clbl[j]=ilbl[j];

      nl=ltd[i][0].nl;
      l=ltd[i][0].l;

      for (j=0;j<nl;j++)
      {
         lt=l[j].type;
         lr=rvt[(int)(lt)];

         k=olbl[(int)(lt)]+clbl[(int)(lt)];
         (*sl).ne[k][nv+l[j].iv[0]]+=1;

         k=olbl[(int)(lr)]+clbl[(int)(lr)];
         (*sl).ne[k][nv+l[j].iv[1]]+=1;

         clbl[(int)(lt)]+=1;
         if (lr!=lt)
            clbl[(int)(lr)]+=1;
      }

      dm=ltd[i][0].dm;
      nv+=(*dm).nv;
   }

   if (th==PHI4)
      lt0=SCALAR;
   else
      lt0=GLUON;

   nv=0;

   for (i=0;i<nd;i++)
   {
      dm=ltd[i][0].dm;
      v=(*dm).v;

      for (j=0;j<(*dm).nv;j++)
      {
         vx=vls.v[v[j][0]]+v[j][1];

         for (k=0;k<(int)(LTYPES);k++)
         {
            if (ilbl[k])
            {
               (*sl).ne[olbl[k]][nv+j]=(*vx).nl[k];

               for (m=1;m<=nlbl[k];m++)
                  (*sl).ne[olbl[k]][nv+j]-=(*sl).ne[olbl[k]+m][nv+j];
            }
         }
      }

      rv=ltd[i][0].rv;

      if ((rv<(*dm).nv)&&(ilbl[(int)(lt0)]))
         (*sl).ne[olbl[(int)(lt0)]][nv+rv]-=1;

      l=(*dm).l;

      for (j=0;j<(*dm).nl;j++)
      {
         lt=l[j].type;
         lr=rvt[(int)(lt)];

         if (ilbl[(int)(lt)])
            (*sl).ne[olbl[(int)(lt)]][nv+l[j].iv[0]]-=1;

         if (ilbl[(int)(lr)])
            (*sl).ne[olbl[(int)(lr)]][nv+l[j].iv[1]]-=1;
      }

      nv+=(*dm).nv;
   }
}


static void clean_up(src_legs_t *sl)
{
   int nt,nv,ns;
   int n,i,j;

   nt=(*sl).nt;

   for (i=0;i<nt;i++)
   {
      nv=(*sl).nv[i];
      ns=0;
      n=0;

      for (j=0;j<nv;j++)
      {
         if ((*sl).ne[i][j]>0)
         {
            ns+=(*sl).ne[i][j];
            (*sl).iv[i][n]=(*sl).iv[i][j];
            (*sl).ne[i][n]=(*sl).ne[i][j];
            n+=1;
         }
      }

      (*sl).nv[i]=n;
      (*sl).ns[i]=ns;
   }
}


static void set_vls(theory_t th,int nd,ltree_t **ltd)
{
   int nv,i,j,k;
   int (*v)[2],order;
   diagram_t *dm;

   order=0;

   for (i=0;i<nd;i++)
   {
      dm=ltd[i][0].dm;
      nv=(*dm).nv;
      v=(*dm).v;

      for (j=0;j<nv;j++)
      {
         k=v[j][0];

         if (k>order)
            order=k;
      }
   }

   vls=vertices(th,order);
}


void get_src_legs(theory_t th,int nd,ltree_t **ltd,src_legs_t *sl)
{
   error(nd<1,1,"get_src_legs [src_legs.c]","Empty diagram list");

   set_vls(th,nd,ltd);
   set_types(th,nd,ltd,sl);
   set_ne(th,nd,ltd,sl);
   clean_up(sl);
}
