
/*******************************************************************************
*
* File pltree.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Drawing of labeled tree diagrams.
*
* The externally accessible functions are
*
*   void draw_ltree0(ltree_t *ltd,vlist *vl)
*     Writes plot commands representing the labeled tree diagram "ltd" to
*     the currently open plot file. An error occurs if the diagram is not
*     a tree diagram. The properties of the vertices in the diagram are
*     retrieved from the vertex list "vl" (see diagrams/vertices.c).
*
*   void draw_ltree1(cnum_t *cn,ltree_t *ltd,vlist *vl)
*     Writes plot commands representing the labeled tree diagram "ltd" and
*     its coefficient "cn" to the currently open plot file. An error occurs
*     if the diagram is not a tree diagram. The properties of the vertices
*     in the diagram are retrieved from the vertex list "vl" (see diagrams/
*     vertices.c).
*
* Notes:
*
* The programs are to be used together with the plot tools provided by the
* module ptools.c. Assuming a plot file has been opened using start_plot(),
* a diagram "dm" can be added to the plot by the commands
*
*  double pos1[2],pos2[2];
*
*  start_bbox();
*  draw_ltree0(ltd,&vl);
*  end_bbox(pos1,pos2);
*
*  start_diagram(pos1,pos2);
*  draw_ltree0(ltd,&vl);
*  end_diagram();
*
* where the vlist_t "vl" is assumed to have been set somewhere before these
* lines.
*
*******************************************************************************/

#define PLTREE_C

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "utils.h"
#include "rvt.h"
#include "plot.h"

#define LINE (SCALE_FACTOR*14.0)
#define VSEP (SCALE_FACTOR*24.0)
#define CSEP (SCALE_FACTOR*28.0)

static int nvs=0,*ns,**nn;
static vlist_t *vls;
static ltype_t **imat;


static void alloc_aux(int nv)
{
   int i;
   ltype_t *ima0;

   if (nvs>0)
   {
      free(ns);
      free(nn);
      free(imat[0]);
      free(imat);
   }

   ns=malloc((nv+nv*nv)*sizeof(*ns));
   nn=malloc(nv*sizeof(*nn));
   imat=malloc(nv*sizeof(*imat));
   ima0=malloc(nv*nv*sizeof(*ima0));

   error((ns==NULL)||(nn==NULL)||(imat==NULL)||(ima0==NULL),1,
         "alloc_aux [pltree.c]","Unable to allocate auxiliary arrays");

   nvs=nv;
   nn[0]=ns+nv;
   imat[0]=ima0;

   for (i=1;i<nv;i++)
   {
      nn[i]=nn[i-1]+nv;
      imat[i]=imat[i-1]+nv;
   }
}


static void set_imat(diagram_t *dm)
{
   int nv,nl,i,j,k;
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;
   l=(*dm).l;

   for (i=0;i<nv;i++)
   {
      for (j=0;j<nv;j++)
         imat[i][j]=LTYPES;
   }

   for (k=0;k<nl;k++)
   {
      i=(*l).iv[0];
      j=(*l).iv[1];

      imat[i][j]=(*l).type;
      imat[j][i]=rvt[(int)((*l).type)];

      l+=1;
   }
}


static int set_nn(diagram_t *dm)
{
   int nv,i,j;

   nv=(*dm).nv;

   if (nv>nvs)
      alloc_aux(nv);

   if (is_tree(dm)==1)
   {
      set_imat(dm);

      for (i=0;i<nv;i++)
      {
         ns[i]=0;

         for (j=0;j<nv;j++)
         {
            if (imat[i][j]!=LTYPES)
            {
               nn[i][ns[i]]=j;
               ns[i]+=1;
            }
         }
      }

      return 1;
   }
   else
      return 0;
}


static int maxnv(int nv,int ivo,int ivn)
{
   int nmx,n,i,iv;

   nmx=0;

   for (i=0;i<ns[ivn];i++)
   {
      iv=nn[ivn][i];

      if (iv!=ivo)
      {
         n=maxnv(nv,ivn,iv);

         if (n>nmx)
            nmx=n;
      }
   }

   return 1+nmx;
}


static int first_vx(diagram_t *dm)
{
   int nv,nmx,n,i,iv;

   nv=(*dm).nv;
   nmx=0;
   iv=0;

   for (i=0;i<nv;i++)
   {
      if (ns[i]<2)
      {
         n=maxnv(nv,i,i);

         if (n>nmx)
         {
            nmx=n;
            iv=i;
         }
      }
   }

   return iv;
}


static int next_vx(int nv,int ivo,int ivn)
{
   int nmx,n,i,iv,ivnn;

   nmx=0;
   iv=ivn;

   for (i=0;i<ns[ivn];i++)
   {
      ivnn=nn[ivn][i];

      if (ivnn!=ivo)
      {
         n=maxnv(nv,ivn,ivnn);

         if (n>nmx)
         {
            nmx=n;
            iv=ivnn;
         }
      }
   }

   return iv;
}


static void set_phi(int ie,int n,double *phi,double *dphi)
{
   double p,dp;

   if (ie==0)
   {
      dp=360.0/(double)(n);
      p=0.0;
   }
   else if (ie==1)
   {
      dp=360.0/(double)(n+1);
      p=dp;
   }
   else if (ie==2)
   {
      dp=360.0/(double)(n+1);
      p=180.0+dp;
   }
   else if (ie==3)
   {
      if (n<=2)
      {
         dp=180.0/(double)(n+1);
         p=dp;
      }
      else
      {
         dp=60.0/(double)(n-1);
         p=60.0;
      }
   }
   else
   {
      if (n<=2)
      {
         dp=180.0/(double)(n+1);
         p=180.0+dp;
      }
      else
      {
         dp=60.0/(double)(n-1);
         p=240.0;
      }
   }

   (*phi)=p;
   (*dphi)=dp;
}


static int line_label(ltree_t *ltd,ltype_t lt,int iv,int nc)
{
   int nl,lbl,i;
   ltype_t lr;
   line_t *l;

   nl=(*ltd).nl;
   l=(*ltd).l;
   lr=rvt[(int)(lt)];
   lbl=0;

   for (i=0;i<nl;i++)
   {
      if ((lt==(*l).type)||(lr==(*l).type))
         lbl+=1;

      if ((lt==(*l).type)&&(iv==(*l).iv[0]))
      {
         if (nc==0)
            return lbl;
         nc-=1;
      }

      if ((lr==(*l).type)&&(iv==(*l).iv[1]))
      {
         if (nc==0)
            return lbl;
         nc-=1;
      }

      l+=1;
   }

   return 0;
}


static void draw_vx(ltree_t *ltd,double theta,int ivo,int ivn);


static void draw_vx0(ltree_t *ltd,double theta,int ivo,int ivn)
{
   int nv,nl,iv,lbl,rv;
   int n,i,j,ie,iw;
   double phi,dphi,vec[2];
   vertex_t *vx;
   diagram_t *dm;

   rv=(*ltd).rv;
   dm=(*ltd).dm;
   nv=(*dm).nv;
   vx=(*vls).v[(*dm).v[ivn][0]]+(*dm).v[ivn][1];

   if (ns[ivn]==2)
   {
      if (ivo==nn[ivn][0])
         iv=nn[ivn][1];
      else
         iv=nn[ivn][0];
   }
   else if ((ns[ivn]==1)&&(ivo==ivn))
      iv=nn[ivn][0];
   else
      iv=ivn;

   n=-(ivo!=ivn)-(ivn!=iv);

   for (i=0;i<(int)(LTYPES);i++)
      n+=(*vx).nl[i];

   if ((ivo==ivn)&&(ivn==iv))
      ie=0;
   else if (ivo==ivn)
      ie=1;
   else if (ivn==iv)
      ie=2;
   else
      ie=3;

   if (ie==3)
      set_phi(ie,(n+1)/2,&phi,&dphi);
   else
      set_phi(ie,n,&phi,&dphi);

   vec[0]=VSEP;
   vec[1]=0.0;

   if (ivn!=iv)
      draw_line(imat[ivn][iv],theta,vec);

   vec[0]=LINE;
   iw=0;

   for (i=0;i<(int)(LTYPES);i++)
   {
      nl=(*vx).nl[i]-(i==(int)(imat[ivn][ivo]))-(i==(int)(imat[ivn][iv]));

      for (j=0;j<nl;j++)
      {
         lbl=line_label(ltd,(ltype_t)(i),ivn,j);

         if ((rv==ivn)&&(lbl==0)&&
             (((ltype_t)(i)==SCALAR)||((ltype_t)(i)==GLUON)))
         {
            draw_rooted_line((ltype_t)(i),theta+phi,vec);
            rv=nv;
         }
         else
            draw_labeled_line((ltype_t)(i),theta+phi,vec,lbl);

         phi+=dphi;
         iw+=1;

         if ((ie==3)&&(iw==((n+1)/2)))
         {
            ie=4;
            set_phi(ie,n/2,&phi,&dphi);
         }
      }
   }

   draw_vertex((*vx).type);

   if (ivn!=iv)
   {
      vec[0]=VSEP;
      move_point(theta,vec);
      draw_vx(ltd,theta,ivn,iv);
   }
}


static void draw_vx1(ltree_t *ltd,double theta,int ivo,int ivn)
{
   int nv,nl,iv,ivnn,lbl,rv;
   int n,i,j,iw;
   double phi,dphi,vec[2],old[2];
   vertex_t *vx;
   diagram_t *dm;

   get_point(old);
   rv=(*ltd).rv;
   dm=(*ltd).dm;
   nv=(*dm).nv;
   vx=(*vls).v[(*dm).v[ivn][0]]+(*dm).v[ivn][1];
   iv=next_vx((*dm).nv,ivo,ivn);

   vec[1]=0.0;
   n=-ns[ivn];

   for (i=0;i<(int)(LTYPES);i++)
      n+=(*vx).nl[i];

   if (n>0)
   {
      set_phi(4,n,&phi,&dphi);
      vec[0]=LINE;

      for (i=0;i<(int)(LTYPES);i++)
      {
         nl=(*vx).nl[i];

         for (j=0;j<ns[ivn];j++)
            nl-=(i==(int)(imat[ivn][nn[ivn][j]]));

         for (j=0;j<nl;j++)
         {
               lbl=line_label(ltd,(ltype_t)(i),ivn,j);

            if ((rv==ivn)&&(lbl==0)&&
                (((ltype_t)(i)==SCALAR)||((ltype_t)(i)==GLUON)))
            {
               draw_rooted_line((ltype_t)(i),theta+phi,vec);
               rv=nv;
            }
            else
               draw_labeled_line((ltype_t)(i),theta+phi,vec,lbl);

            phi+=dphi;
         }
      }

      set_phi(3,ns[ivn]-2,&phi,&dphi);
      vec[0]=VSEP;

      for (i=0;i<ns[ivn];i++)
      {
         ivnn=nn[ivn][i];

         if ((ivnn!=ivo)&&(ivnn!=iv))
         {
            draw_line(imat[ivn][ivnn],theta+phi,vec);
            move_point(theta+phi,vec);
            draw_vx(ltd,theta+phi,ivn,ivnn);
            set_point(old);
            phi+=dphi;
         }
      }
   }
   else
   {
      n=ns[ivn]-2;
      vec[0]=VSEP;
      set_phi(3,(n+1)/2,&phi,&dphi);
      iw=0;

      for (i=0;i<ns[ivn];i++)
      {
         ivnn=nn[ivn][i];

         if ((ivnn!=ivo)&&(ivnn!=iv))
         {
            draw_line(imat[ivn][ivnn],theta+phi,vec);
            move_point(theta+phi,vec);
            draw_vx(ltd,theta+phi,ivn,ivnn);
            set_point(old);
            phi+=dphi;
            iw+=1;

            if (iw==((n+1)/2))
               set_phi(4,n/2,&phi,&dphi);
         }
      }
   }

   draw_line(imat[ivn][iv],theta,vec);
   draw_vertex((*vx).type);
   move_point(theta,vec);
   draw_vx(ltd,theta,ivn,iv);
}


static void draw_vx(ltree_t *ltd,double theta,int ivo,int ivn)
{
   if (ns[ivn]<3)
      draw_vx0(ltd,theta,ivo,ivn);
   else
      draw_vx1(ltd,theta,ivo,ivn);
}


void draw_ltree0(ltree_t *ltd,vlist_t *vl)
{
   int iv;
   diagram_t *dm;

   dm=(*ltd).dm;

   error(set_nn(dm)==0,1,"draw_tree [pltree.c]",
         "Diagram is not a tree diagram");

   vls=vl;
   iv=first_vx(dm);
   draw_vx(ltd,0.0,iv,iv);
}


void draw_ltree1(cnum_t *cn,ltree_t *ltd,vlist_t *vl)
{
   double vec[2];

   draw_coefficient(cn);
   vec[0]=CSEP;
   vec[1]=0.0;
   move_point(0.0,vec);
   draw_ltree0(ltd,vl);
}
