
/*******************************************************************************
*
* File ltree.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for labeled tree diagrams.
*
* The externally accessible functions are
*
*   ltree_t *alloc_ltd(int nl)
*     Returns the pointer to a newly allocated labeled tree diagram
*     structure with "nl" additional lines. The array of lines is
*     allocated and initialized (integers to 0, line types to LTYPES).
*     The pointer to the tree diagram in the structure is set to NULL.
*
*   void free_ltd(ltree_t *td)
*     Frees the line array in the structure "td" and then the structure
*     itself. The tree diagram pointed to in the structure is not freed.
*
*   void copy_ltd(ltree_t *td1,ltree_t *td2)
*     Copies the labeled tree diagram td1 to td2. The lines are copied
*     one by one, but only the pointer to the tree diagram rather than
*     the diagram itself is copied.
*
*   void ltd2dgm(ltree_t *td,diagram_t *dm)
*     Assigns the contracted diagram defined by the labeled tree diagram
*     "td" to the diagram structure "dm".
*
* Notes:
*
* Labeled tree diagrams are tree diagrams with a list of additional lines
* of type line_t. These lines must be such that
*
*  - Both endpoints of the lines are vertices of the tree diagram.
*
*  - Each fermion line closes a fermion path in the tree diagram
*    or goes from a vertex to itself.
*
* With respect to a particular theory, labeled tree diagrams are further
* required to be such that the contracted diagram (the diagram obtained by
* adding all lines to the tree diagram) is a possible diagram.
*
* The list of additional lines may be thought of as a way of specifying
* contractions of external lines of the tree diagram. Labeled tree diagrams
* may thus be graphically represented by drawing the tree diagram and by
* labeling the external lines that get contracted with the index of the
* corresponding line in the array of additional lines of equal type.
*
* Labeled tree diagrams that lead to the same contracted diagram are not
* distinguished. The ordering of the lines in the list of additional lines
* can thus be chosen arbitrarily.
*
* The elements of an ltree_t structure are:
*
*  nl           Number of additional lines.
*
*  l            Array l[0],..,l[nl-1] of the additional lines.
*
*  dm           Pointer to the structure that describes the tree
*               diagram.
*
* A further element
*
*  rv           Index of the vertex where the root line is attached.
*               rv is set to the number nv of vertices if the diagram
*               is unrooted.
*
* is included in the structure so as to be able to describe rooted labeled
* tree diagrams with the same data type.
*
*******************************************************************************/

#define LTREE_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "trees.h"


ltree_t *alloc_ltd(int nl)
{
   int i;
   line_t *l;
   ltree_t *td;

   error(nl<0,1,"alloc_ltd [ltree.c]","Argument is out of range");

   td=malloc(1*sizeof(*td));
   if (nl>0)
      l=malloc(nl*sizeof(*l));
   else
      l=NULL;

   error((td==NULL)||((nl>0)&&(l==NULL)),1,
         "alloc_ltd [ltree.c]","Unable to allocate structure");

   (*td).nl=nl;
   (*td).rv=0;
   (*td).l=l;
   (*td).dm=NULL;

   for (i=0;i<nl;i++)
   {
      (*l).type=LTYPES;
      (*l).iv[0]=0;
      (*l).iv[1]=0;
      l+=1;
   }

   return td;
}


void free_ltd(ltree_t *td)
{
   if (td!=NULL)
   {
      if ((*td).nl>0)
         free((*td).l);

      free(td);
   }
}


void copy_ltd(ltree_t *td1,ltree_t *td2)
{
   int nl,i;
   line_t *l1,*l2;

   nl=(*td1).nl;

   error(nl>(*td2).nl,1,"copy_ltd [ltree.c]","Improper arguments");

   (*td2).nl=nl;
   l1=(*td1).l;
   l2=(*td2).l;

   for (i=0;i<nl;i++)
   {
      (*l2).type=(*l1).type;
      (*l2).iv[0]=(*l1).iv[0];
      (*l2).iv[1]=(*l1).iv[1];

      l1+=1;
      l2+=1;
   }

   (*td2).rv=(*td1).rv;
   (*td2).dm=(*td1).dm;
}


void ltd2dgm(ltree_t *td,diagram_t *dm)
{
   int nl,i;
   line_t *l1,*l2;
   diagram_t *tdm;

   nl=(*td).nl;
   tdm=(*td).dm;

   error((tdm==NULL)||((*dm).nl<(nl+(*tdm).nl)),1,"ltd2dgm [ltree.c]",
         "Improper arguments");

   copy_dgm(tdm,dm);
   l1=(*td).l;
   l2=(*dm).l+(*dm).nl;

   for (i=0;i<nl;i++)
   {
      (*l2).type=(*l1).type;
      (*l2).iv[0]=(*l1).iv[0];
      (*l2).iv[1]=(*l1).iv[1];

      l1+=1;
      l2+=1;
   }

   (*dm).nl+=nl;
}
