
/*******************************************************************************
*
* File ltree_gen.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of labeled tree diagrams.
*
* The externally accessible functions are
*
*   int ltrees(theory_t th,int order)
*     Generates all labeled tree diagrams in the theory with label "th"
*     up to and including the specified order in the coupling. The program
*     returns the total number of these diagrams and stores them in static
*     arrays that can be accessed using other functions in this module. An
*     error occurs if trees() [tree.c] has not been called before or with
*     a lower value of the parameter "order".
*
*   int nlp_max(theory_t th,int order)
*     Returns the maximal number of pairs of labeled external lines of the
*     tree diagrams in the theory with label "th" at the specified order in
*     the coupling. An error occurs if ltrees() has not been called before
*     or with a lower value of the parameter "order".
*
*   int no_of_ltrees(theory_t th,int order,int nlp)
*     Returns the number of labeled tree diagrams in the theory with label
*     "th" at the specified order in the coupling and "nlp" pairs of labeled
*     external lines. An error occurs if ltrees() has not been called before
*     or with a lower value of the parameter "order".
*
*   tlist_t *select_ltrees(theory_t th,int order,int nlp)
*     Returns the pointer to the the internally stored tlist_t structure
*     containing all labeled tree diagrams in the theory with label "th" at
*     the specified order in the coupling and "nlp" pairs of labeled external
*     lines. The program returns NULL if there are no such diagrams. An error
*     occurs if ltrees() has not been called before or with a lower value of
*     the parameter "order".
*
*   void set_root(theory_t th,int is)
*     Sets the root line of the labeled tree diagrams in the theory with
*     label "th" up to the available order in the coupling. A line at the
*     first (is=0) or last (is!=0) suitable vertex is chosen. This program
*     may only be called once for each theory.
*
* Notes:
*
* Apart from being labeled tree diagrams (see ltree.c), the diagrams listed
* by ltrees() are required to have no external fermion lines and two or more
* external bosonic lines. Furthermore, cutting a gluon line of the associated
* contracted diagram may not produce a disconnected diagram with the cut line
* being the only external line.
*
* The type tlist_t is described at the top of the file tdlist.c. For a given
* theory, the list of tree diagrams constructed by ltrees() cannot be modified
* by calling ltrees() a second time. Repeated calls of ltrees() with the same
* values of the arguments are permissible and have no effect on the stored list
* of labeled tree diagrams.
*
* The basic tree diagrams in the list generated by ltrees() are copied from
* the list generated by trees() and thus have their lines and vertices ordered
* by reorder_dgm() [diagrams/order.c]. In the labeled-tree structures, the
* added lines are of type GHOST, QUARK, SCALAR or GLUON and appear in this
* order.
*
* Depending on whether the flag "is" vanishes or not, the program set_root()
* searches for the first vertex with the lowest or highest rank, respectively,
* which has an unlabeled external boson line.
*
*******************************************************************************/

#define LTREE_GEN_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "permutations.h"
#include "rvt.h"
#include "trees.h"
#include "contract.h"

typedef struct
{
   int order,rooted;
   int *nlp;
   dlist_t ***dl;
   tlist_t ***tl;
} tsets_t;

static int nvs=0,nls=0,nas=0;
static int *ivs,*rkw,(*nnv)[2],(*lnl)[2];
static diagram_t *dmw;
static ltree_t *ltw;
static vlist_t vls;
static src_legs_t sl;
static tsets_t trs[(int)(THEORIES)]={{0,0,NULL,NULL,NULL},
                                     {0,0,NULL,NULL,NULL},
                                     {0,0,NULL,NULL,NULL},
                                     {0,0,NULL,NULL,NULL},
                                     {0,0,NULL,NULL,NULL}};


static void alloc_aux(theory_t th,int order)
{
   int nv,nl,ne,na;

   tree_lines(th,order,&nl,&ne,&na);
   na=na/2;
   nv=nl+1;
   nl=nl+na;

   if ((nv>nvs)||(nl>nls)||(na>nas))
   {
      if (nvs>0)
      {
         free(ivs);
         free(nnv);
         free_dgm(dmw);
         free_ltd(ltw);
      }
      else
         init_src_legs(&sl);

      if (nv>nvs)
         nvs=nv;
      if (nl>nls)
         nls=nl;
      if (na>nas)
         nas=na;

      ivs=malloc(2*nvs*sizeof(*ivs));
      nnv=malloc((nvs+nas)*sizeof(*nnv));

      error((ivs==NULL)||(nnv==NULL),1,
            "alloc_aux [ltree_gen.c]","Unable to allocate auxiliary arrays");

      rkw=ivs+nvs;
      lnl=nnv+nvs;

      dmw=alloc_dgm(nvs,nls);
      ltw=alloc_ltd(nas);
   }
}


static void alloc_trs(theory_t th,int order)
{
   int n,i,j,*nlp;
   dlist_t ***dl,**dll;
   tlist_t ***tl,**tll;
   tsets_t *ts;

   nlp=malloc((order+1)*sizeof(*nlp));
   error(nlp==NULL,1,"alloc_trs [ltree_gen.c]",
         "Unable to allocate labeled tree set structure");

   nlp[0]=0;
   n=order*(nas+1);

   for (i=1;i<=order;i++)
      nlp[i]=nas+1;

   dl=malloc((order+1)*sizeof(*dl));
   dll=malloc(n*sizeof(*dll));

   tl=malloc((order+1)*sizeof(*tl));
   tll=malloc(n*sizeof(*tll));

   error((dl==NULL)||(dll==NULL)||(tl==NULL)||(tll==NULL),2,
         "alloc_trs [ltree_gen.c]",
         "Unable to allocate labeled tree set structure");

   dl[0]=NULL;
   tl[0]=NULL;

   for (i=1;i<=order;i++)
   {
      dl[i]=dll;
      tl[i]=tll;

      for (j=0;j<nlp[i];j++)
      {
         dll[j]=alloc_dlist(32,1);
         tll[j]=alloc_tlist(32);
      }

      dll+=nlp[i];
      tll+=nlp[i];
   }

   ts=trs+(int)(th);
   (*ts).order=order;
   (*ts).rooted=0;
   (*ts).nlp=nlp;
   (*ts).dl=dl;
   (*ts).tl=tl;
}


static void set_nnv(diagram_t *dm,ltype_t lt)
{
   int nv,nl,iv0,iv1;
   int i,(*v)[2];
   ltype_t lr;
   line_t *l;
   vertex_t *vx;

   lr=rvt[(int)(lt)];
   nv=(*dm).nv;
   v=(*dm).v;

   for (i=0;i<nv;i++)
   {
      vx=vls.v[v[i][0]]+v[i][1];

      if ((*vx).nl[(int)(lt)]!=0)
      {
         ivs[i]=1;
         nnv[i][0]=nv;
         nnv[i][1]=nv;
      }
      else
         ivs[i]=0;
   }

   nl=(*dm).nl;
   l=(*dm).l;

   for (i=0;i<nl;i++)
   {
      if (l[i].type==lt)
      {
         iv0=l[i].iv[0];
         iv1=l[i].iv[1];

         nnv[iv0][1]=iv1;
         nnv[iv1][0]=iv0;
      }
      else if (l[i].type==lr)
      {
         iv0=l[i].iv[0];
         iv1=l[i].iv[1];

         nnv[iv0][0]=iv1;
         nnv[iv1][1]=iv0;
      }
   }
}


static void dgm2ltw(diagram_t *dm)
{
   int nv,nl,i,j,k;
   ltype_t lt;
   line_t *l;

   nv=(*dm).nv;
   nl=0;
   l=(*ltw).l;

   for (i=0;i<2;i++)
   {
      if (i==0)
         lt=GHOST;
      else
         lt=QUARK;

      set_nnv(dm,lt);

      for (j=0;j<nv;j++)
      {
         if ((ivs[j]==1)&&(nnv[j][0]==nv))
         {
            k=j;

            while (nnv[k][1]!=nv)
            {
               k=nnv[k][1];
               ivs[k]=0;
            }

            l[nl].type=lt;
            l[nl].iv[0]=k;
            l[nl].iv[1]=j;

            nl+=1;
         }
      }
   }

   (*ltw).nl=nl;
   (*ltw).rv=nv;
   (*ltw).dm=dm;
}


static void add_ltw(ltype_t lt,dlist_t *dl,tlist_t *tl)
{
   int n1pt;
   ltree_t *ltn;

   (*dmw).nv=nvs;
   (*dmw).nl=nls;
   ltd2dgm(ltw,dmw);

   if ((nlext(dmw,&vls,&n1pt)<2)||(has_tadpole(dmw,lt,&vls)))
      return;

   reorder_dgm(dmw,rkw);

   if (insert_copy(dmw,rkw,dl)==1)
   {
      ltn=alloc_ltd((*ltw).nl);
      copy_ltd(ltw,ltn);
      add_ltd(ltn,tl);
   }
}


void contract_fermions(theory_t th,int order)
{
   int nbr,nl;
   int n,i,j,k;
   ltype_t lt;
   diagram_t **dma;
   dlist_t *dl;
   tsets_t *ts;

   if (th==PHI4)
      lt=SCALAR;
   else
      lt=GLUON;

   ts=trs+(int)(th);

   for (i=1;i<=order;i++)
   {
      nbr=nbr_max(th,i);

      for (j=1;j<=nbr;j++)
      {
         dl=select_trees(th,i,j);

         if (dl!=NULL)
         {
            n=(*dl).n;
            dma=(*dl).dma;

            for (k=0;k<n;k++)
            {
               dgm2ltw(dma[k]);
               nl=(*ltw).nl;
               add_ltw(lt,(*ts).dl[i][nl],(*ts).tl[i][nl]);
            }

            free(dl);
         }
      }
   }
}


static void contract_bosons(theory_t th,int order)
{
   int nl,i,j,k,m;
   int n,*nlp;
   long np,npm,nc;
   ltype_t lt;
   line_t *l;
   ltree_t **lta;
   tlist_t *tl;
   tsets_t *ts;

   if (th==PHI4)
      lt=SCALAR;
   else
      lt=GLUON;

   ts=trs+(int)(th);
   nlp=(*ts).nlp;

   for (i=1;i<=order;i++)
   {
      for (j=(nlp[i]-2);j>=0;j--)
      {
         tl=(*ts).tl[i][j];
         n=(*tl).n;
         lta=(*tl).lta;

         for (k=0;k<n;k++)
         {
            get_src_legs(th,1,lta,&sl);

            for (nl=1;(2*nl)<=(sl.ns[0]-2);nl++)
            {
               npm=init_lnlist(1,sl.nv,&nl,sl.iv,sl.ne);
               np=0L;

               while (np<npm)
               {
                  copy_ltd(*lta,ltw);
                  np=next_lnlist(lnl,&nc);
                  l=(*ltw).l+j;

                  for (m=0;m<nl;m++)
                  {
                     (*l).type=lt;
                     (*l).iv[0]=lnl[m][0];
                     (*l).iv[1]=lnl[m][1];
                     l+=1;
                  }

                  (*ltw).nl=j+nl;
                  add_ltw(lt,(*ts).dl[i][j+nl],(*ts).tl[i][j+nl]);
               }

               stop_lnlist();
            }

            lta+=1;
         }
      }
   }
}


static void reset_nlp(theory_t th,int order)
{
   int n,i,j;
   int *nlp;
   tlist_t *tl;
   tsets_t *ts;

   ts=trs+(int)(th);
   nlp=(*ts).nlp;

   for (i=1;i<=order;i++)
   {
      n=0;

      for (j=1;j<nlp[i];j++)
      {
         tl=(*ts).tl[i][j];

         if ((*tl).n>0)
            n=j;
      }

      for (j=n+1;j<nlp[i];j++)
      {
         tl=(*ts).tl[i][j];
         free_tlist(tl);
         (*ts).tl[i][j]=NULL;
      }

      nlp[i]=n+1;
   }
}


static int nall_dgm(theory_t th,int order)
{
   int nall,i,j;
   int *nlp;
   tlist_t *tl;
   tsets_t *ts;

   ts=trs+(int)(th);
   nlp=(*ts).nlp;
   nall=0;

   for (i=1;i<=order;i++)
   {
      for (j=0;j<nlp[i];j++)
      {
         tl=(*ts).tl[i][j];
         nall+=(*tl).n;
      }
   }

   return nall;
}


int ltrees(theory_t th,int order)
{
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"ltrees [ltree_gen.c]",
         "Arguments are out of range");
   ts=trs+(int)(th);

   if ((order>0)&&((*ts).order==0))
   {
      vls=vertices(th,order);
      alloc_aux(th,order);
      alloc_trs(th,order);
      contract_fermions(th,order);
      contract_bosons(th,order);
      reset_nlp(th,order);
   }
   else
      error(order>(*ts).order,1,"ltrees [ltree_gen.c]",
            "Attempt to extend the labeled-tree list");

   if (order>0)
      return nall_dgm(th,order);
   else
      return 0;
}


int nlp_max(theory_t th,int order)
{
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"nlp_max [ltree_gen.c]",
         "Arguments are out of range");

   if (order==0)
      return 0;

   ts=trs+(int)(th);
   error(order>(*ts).order,1,"nlp_max [ltree_gen.c]",
         "Requested diagrams are not included in current list");

   return (*ts).nlp[order]-1;
}


int no_of_ltrees(theory_t th,int order,int nlp)
{
   tlist_t *tl;
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"no_of_ltrees [ltree_gen.c]",
         "Arguments are out of range");

   if (order==0)
      return 0;

   ts=trs+(int)(th);
   error(order>(*ts).order,1,"no_of_ltrees [ltree_gen.c]",
         "Requested diagrams are not included in current list");

   error((nlp<0)||(nlp>=(*ts).nlp[order]),1,"no_of_ltrees [ltree_gen.c]",
         "Parameter nlp is out of range");

   tl=(*ts).tl[order][nlp];

   return (*tl).n;
}


tlist_t *select_ltrees(theory_t th,int order,int nlp)
{
   tlist_t *tl;
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"select_ltrees [ltree_gen.c]",
         "Arguments are out of range");

   if (order==0)
      return NULL;

   ts=trs+(int)(th);
   error(order>(*ts).order,1,"select_ltrees [ltree_gen.c]",
         "Requested diagrams are not included in current list");

   error((nlp<0)||(nlp>=(*ts).nlp[order]),1,"select_ltrees [ltree_gen.c]",
         "Parameter nlp is out of range");

   tl=(*ts).tl[order][nlp];

   if ((*tl).n==0)
      return NULL;
   else
      return tl;
}


static void set_rv(ltype_t lt,ltree_t *ltd,int is)
{
   int nv,nl,(*v)[2];
   int n,i,j,k;
   vertex_t *vx;
   line_t *l;
   diagram_t *dm;

   dm=(*ltd).dm;
   nv=(*dm).nv;
   v=(*dm).v;

   for (i=0;i<nv;i++)
   {
      if (is==0)
         k=i;
      else
         k=nv-i-1;

      vx=vls.v[v[k][0]]+v[k][1];
      n=(*vx).nl[(int)(lt)];

      nl=(*ltd).nl;
      l=(*ltd).l;

      for (j=0;(n>0)&&(j<nl);j++)
      {
         if (l[j].type==lt)
         {
            if (l[j].iv[0]==k)
               n-=1;
            if (l[j].iv[1]==k)
               n-=1;
         }
      }

      nl=(*dm).nl;
      l=(*dm).l;

      for (j=0;(n>0)&&(j<nl);j++)
      {
         if (l[j].type==lt)
         {
            if (l[j].iv[0]==k)
               n-=1;
            if (l[j].iv[1]==k)
               n-=1;
         }
      }

      if (n>0)
      {
         (*ltd).rv=k;
         return;
      }
   }

   error(1,1,"set_rv [ltree_gen.c]","Unable to assign root line");
}


void set_root(theory_t th,int is)
{
   int order,*nlp;
   int n,i,j,k;
   ltype_t lt;
   ltree_t **lta;
   tlist_t *tl;
   tsets_t *ts;

   error(th==THEORIES,1,"set_root [ltree_gen.c]",
         "Unknown theory");

   ts=trs+(int)(th);

   error((*ts).rooted!=0,1,"set_root [ltree_gen.c]",
         "Diagrams are already rooted");

   (*ts).rooted=1;
   order=(*ts).order;
   nlp=(*ts).nlp;
   vls=vertices(th,order);

   if (th==PHI4)
      lt=SCALAR;
   else
      lt=GLUON;

   for (i=1;i<=order;i++)
   {
      for (j=0;j<nlp[i];j++)
      {
         tl=(*ts).tl[i][j];
         n=(*tl).n;
         lta=(*tl).lta;

         for (k=0;k<n;k++)
            set_rv(lt,lta[k],is);
      }
   }
}
