
/*******************************************************************************
*
* File tree_gen.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of the basic tree diagrams.
*
* The externally accessible functions are
*
*   int trees(theory_t th,int order)
*     Generates all tree diagrams in the theory with label "th" up to and
*     including the specified order in the coupling. The program returns
*     the total number of trees and stores the diagrams in static arrays
*     that can be accessed using other functions in this module.
*
*   int nbr_max(theory_t th,int order)
*     Returns the maximal number of branches of the tree diagrams in the
*     theory with label "th" at the specified order in the coupling.
*
*   int no_of_trees(theory_t th,int order,int nbr)
*     Returns the number of tree diagrams in the theory with label "th"
*     at the specified order in the coupling with "nbr" branches.
*
*   dlist_t *select_trees(theory_t th,int order,int nbr)
*     Returns the pointer to the internally stored structure containing
*     all tree diagrams in the theory with label "th" at the specified
*     order in the coupling with "nbr" branches. The program returns NULL
*     if there are no such diagrams.
*
*   void tree_lines(theory_t th,int order,int *nlall,int *nevtx,int *neall)
*     This program runs through all tree diagrams in the theory with label
*     "th" up to the specified order in the coupling and determines the
*     maximal number "nlall" of internal lines of the diagrams, the maximal
*     number "nevtx" of legs of the vertices and the maximal number "neall"
*     of external lines of the diagrams.
*
* Notes:
*
* For a given theory, the list of tree diagrams constructed by trees() cannot
* be modified by calling trees() a second time. Repeated calls of trees() with
* the same values of the arguments are permissible and have no effect on the
* stored list of tree diagrams.
*
* All other programs assume that trees() has previously been called for the
* specified theory and with an equal or larger value of the argument "order".
* An error occurs if this is not the case.
*
* The type dlist_t is described at the top of the file dmlist.c. A tree
* diagram is a connected diagram without loops and 1-point vertices.
* Diagrams that differ by the labeling of the vertices and lines are not
* distinguished.
*
* Tree diagrams can have a number nbr of branches, the simplest ones having
* nbr=1 branch. Diagrams with nbr=2 branches have a second branch attached to
* one of the middle vertices of the first branch, those with nbr=3 a further
* branch attached to a middle vertex of a diagram with 2 branches, and so on
* (the term "middle vertex" stands for any vertex, where at least 2 internal
* lines end).
*
* All diagrams stored have their vertices ranked and ordered by reorder_dgm()
* [diagrams/order.c]. Coefficient arrays are not allocated and rank arrays are
* used for internal purposes only, i.e. they are not included in the structures
* returned by select_trees().
*
*******************************************************************************/

#define TREE_GEN_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "permutations.h"
#include "rvt.h"
#include "trees.h"

typedef struct
{
   int order;
   int *nat;
   dlist_t ***dl;
} tsets_t;

static int nvm=0,nlm;
static int **nbrs,**ivs,*rkw;
static vlist_t vls;
static branch_t *brw;
static diagram_t *dmw;
static dlist_t ***brs;
static tsets_t trs[(int)(THEORIES)]={{0,NULL,NULL},{0,NULL,NULL},{0,NULL,NULL},
                                     {0,NULL,NULL},{0,NULL,NULL}};


static void alloc_brs(int order)
{
   int nt,i,j;
   int *pn;
   dlist_t **pb;

   if (nvm>0)
   {
      for (i=0;i<=(nvm-2);i++)
      {
         for (j=0;j<(int)(LTYPES);j++)
         {
            if (nbrs[i][j]>0)
               deep_free_dlist(brs[i][j]);
         }
      }

      free(nbrs[0]);
      free(nbrs);
      free(brs[0]);
      free(brs);
      free_branch(brw);
      free_dgm(dmw);
   }

   nvm=order+2;
   nlm=nvm-1;
   nt=(int)(LTYPES);

   nbrs=malloc((order+1+nvm)*sizeof(*nbrs));
   pn=malloc(((order+1)*nt+nvm*nt+nvm)*sizeof(*pn));
   brs=malloc((order+1)*sizeof(*brs));
   pb=malloc((order+1)*nt*sizeof(*pb));

   error((nbrs==NULL)||(pn==NULL)||(brs==NULL)||(pb==NULL),1,
         "alloc_brs [tree_gen.c]","Unable to allocate auxiliary arrays");

   for (i=0;i<=order;i++)
   {
      nbrs[i]=pn;
      brs[i]=pb;

      for (j=0;j<nt;j++)
      {
         pn[j]=0;
         pb[j]=NULL;
      }

      pn+=nt;
      pb+=nt;
   }

   ivs=nbrs+order+1;

   for (i=0;i<nvm;i++)
   {
      ivs[i]=pn;

      for (j=0;j<nt;j++)
         pn[j]=0;

      pn+=nt;
   }

   rkw=pn;

   for (i=0;i<nvm;i++)
      rkw[i]=0;

   brw=alloc_branch(nvm);
   dmw=alloc_dgm(nvm,nlm);
}


static void set_brs(theory_t th,int order)
{
   int nbr,ibr,nv,i,j;
   diagram_t *dmn;

   alloc_brs(order);
   nbr=branches(th,order);

   for (ibr=0;ibr<nbr;ibr++)
   {
      (*brw).nv=nvm;
      get_branch(th,ibr,brw);

      if ((*brw).v[0][0]==0)
      {
         nv=(*brw).nv;

         if ((nv>1)&&((*brw).v[nv-1][0]!=0))
         {
            i=get_order(th,ibr);
            j=(int)((*brw).l[0]);
            nbrs[i][j]+=1;
         }
      }
   }

   for (i=0;i<=order;i++)
   {
      for (j=0;j<(int)(LTYPES);j++)
      {
         if (nbrs[i][j]>0)
            brs[i][j]=alloc_dlist(nbrs[i][j],0);
      }
   }

   for (ibr=0;ibr<nbr;ibr++)
   {
      (*brw).nv=nvm;
      get_branch(th,ibr,brw);

      if ((*brw).v[0][0]==0)
      {
         nv=(*brw).nv;

         if ((nv>1)&&((*brw).v[nv-1][0]!=0))
         {
            (*dmw).nv=nvm;
            (*dmw).nl=nlm;
            branch2dgm(brw,dmw);

            dmn=alloc_dgm((*dmw).nv,(*dmw).nl);
            copy_dgm(dmw,dmn);

            i=get_order(th,ibr);
            j=(int)((*brw).l[0]);
            add_dgm(dmn,brs[i][j]);
         }
      }
   }
}


static void alloc_trs(theory_t th,int order)
{
   int i,j,n,*nat;
   dlist_t ***dl,**dll;
   tsets_t *ts;

   nat=malloc((order+1)*sizeof(*nat));
   error(nat==NULL,1,"alloc_trs [tree_gen.c]",
         "Unable to allocate tree set structure");
   n=0;

   for (i=0;i<=order;i++)
   {
      if (i==0)
         nat[i]=0;
      else if (i<=3)
         nat[i]=1;
      else
         nat[i]=i-2;

      n+=nat[i];
   }

   dl=malloc((order+1)*sizeof(*dl));
   dll=malloc(n*sizeof(*dll));
   error((dl==NULL)||(dll==NULL),2,"alloc_trs [tree_gen.c]",
         "Unable to allocate tree set structure");
   dl[0]=NULL;

   for (i=1;i<=order;i++)
   {
      dl[i]=dll;
      dll[0]=alloc_dlist(32,0);

      for (j=1;j<nat[i];j++)
         dll[j]=alloc_dlist(32,1);

      dll+=nat[i];
   }

   ts=trs+(int)(th);
   (*ts).order=order;
   (*ts).nat=nat;
   (*ts).dl=dl;
}


static void first_trees(theory_t th,int order)
{
   int nbr,ibr,ibs,nv,i;
   diagram_t *dmn;
   tsets_t *ts;

   alloc_trs(th,order);
   ts=trs+(int)(th);
   nbr=branches(th,order);

   for (ibr=0;ibr<nbr;ibr++)
   {
      (*brw).nv=nvm;
      get_branch(th,ibr,brw);

      if ((*brw).v[0][0]!=0)
      {
         nv=(*brw).nv;
         ibs=inverse_branch(th,ibr);

         if ((ibs>=ibr)&&((*brw).v[nv-1][0]!=0))
         {
            (*dmw).nv=nvm;
            (*dmw).nl=nlm;
            branch2dgm(brw,dmw);
            reorder_dgm(dmw,rkw);

            dmn=alloc_dgm((*dmw).nv,(*dmw).nl);
            copy_dgm(dmw,dmn);

            i=get_order(th,ibr);
            add_dgm(dmn,(*ts).dl[i][0]);
         }
      }
   }
}


static int set_ivs(diagram_t *dm)
{
   int nv,nl,i,j,n,is;
   int (*v)[2];
   line_t *l;
   vertex_t *vx;

   nv=(*dm).nv;
   nl=(*dm).nl;
   v=(*dm).v;
   l=(*dm).l;

   for (i=0;i<nv;i++)
   {
      for (j=0;j<(int)(LTYPES);j++)
         ivs[i][j]=0;
   }

   for (i=0;i<nl;i++)
   {
      ivs[(*l).iv[0]][(int)((*l).type)]+=1;
      ivs[(*l).iv[1]][(int)(rvt[(int)((*l).type)])]+=1;
      l+=1;
   }

   is=0;

   for (i=0;i<nv;i++)
   {
      n=0;

      for (j=0;j<(int)(LTYPES);j++)
         n+=ivs[i][j];

      if (n<2)
      {
         for (j=0;j<(int)(LTYPES);j++)
            ivs[i][j]=0;
      }
      else
      {
         vx=vls.v[v[i][0]]+v[i][1];

         for (j=0;j<(int)(LTYPES);j++)
         {
            ivs[i][j]=(*vx).nl[j]-ivs[i][j];
            is|=(ivs[i][j]>0);
         }
      }
   }

   return is;
}


static void attach_branch(int iv,diagram_t *dm,diagram_t *br)
{
   int nvn,nln,nv,nl,i;
   int (*v1)[2],(*v2)[2];
   line_t *l1,*l2;

   nvn=(*dm).nv+(*br).nv-1;
   nln=(*dm).nl+(*br).nl;

   (*dmw).nv=nvn;
   (*dmw).nl=nln;

   nv=(*dm).nv;
   v1=(*dm).v;
   v2=(*dmw).v;

   for (i=0;i<nvn;i++)
   {
      if (i==nv)
         v1=(*br).v+1;

      (*v2)[0]=(*v1)[0];
      (*v2)[1]=(*v1)[1];

      v1+=1;
      v2+=1;
   }

   nl=(*dm).nl;
   l1=(*dm).l;
   l2=(*dmw).l;

   for (i=0;i<nl;i++)
   {
      (*l2).type=(*l1).type;
      (*l2).iv[0]=(*l1).iv[0];
      (*l2).iv[1]=(*l1).iv[1];

      l1+=1;
      l2+=1;
   }

   nl=(*br).nl;
   l1=(*br).l;

   for (i=0;i<nl;i++)
   {
      (*l2).type=(*l1).type;
      if (i>0)
         (*l2).iv[0]=nv+i-1;
      else
         (*l2).iv[0]=iv;
      (*l2).iv[1]=nv+i;

      l1+=1;
      l2+=1;
   }
}


static int add_branches(int ib,diagram_t *dm,dlist_t *dl)
{
   int nv,nt,n;
   int is,iv,il,k;
   diagram_t **dma;
   dlist_t *brl;

   if (set_ivs(dm)==0)
      return 0;

   nv=(*dm).nv;
   nt=(int)(LTYPES);
   is=0;

   for (iv=0;iv<nv;iv++)
   {
      for (il=0;il<nt;il++)
      {
         if (ivs[iv][il]>0)
         {
            brl=brs[ib][il];
            n=(*brl).n;
            dma=(*brl).dma;

            for (k=0;k<n;k++)
            {
               attach_branch(iv,dm,dma[k]);
               reorder_dgm(dmw,rkw);
               is|=insert_copy(dmw,rkw,dl);
            }
         }
      }
   }

   return is;
}


static int next_trees(int ntree,theory_t th,int order)
{
   int io,ib,in;
   int i,is,*nat;
   dlist_t ***dl,*dlo,*dln;
   diagram_t **dmo;

   nat=trs[(int)(th)].nat;
   dl=trs[(int)(th)].dl;
   is=0;

   for (io=1;io<order;io++)
   {
      for (ib=1;ib<=(order-io);ib++)
      {
         in=io+ib;

         if ((ntree<nat[io])&&((ntree+1)<nat[in]))
         {
            dlo=dl[io][ntree];
            dln=dl[in][ntree+1];
            dmo=(*dlo).dma;

            for (i=0;i<(*dlo).n;i++)
               is|=add_branches(ib,dmo[i],dln);
         }
      }
   }

   return is;
}


static void reset_nat(theory_t th,int order)
{
   int n,i,j;
   int *nat;
   dlist_t *dl;
   tsets_t *ts;

   ts=trs+(int)(th);
   nat=(*ts).nat;

   for (i=1;i<=order;i++)
   {
      n=1;

      for (j=0;j<nat[i];j++)
      {
         dl=(*ts).dl[i][j];

         if ((*dl).n>0)
            n=j+1;
      }

      nat[i]=n;
   }
}


static int nall_dgm(theory_t th,int order)
{
   int nall,i,j;
   int *nat;
   dlist_t *dl;
   tsets_t *ts;

   ts=trs+(int)(th);
   nat=(*ts).nat;
   nall=0;

   for (i=1;i<=order;i++)
   {
      for (j=0;j<nat[i];j++)
      {
         dl=(*ts).dl[i][j];
         nall+=(*dl).n;
      }
   }

   return nall;
}


int trees(theory_t th,int order)
{
   int ntree;
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"trees [tree_gen.c]",
         "Arguments are out of range");
   ts=trs+(int)(th);

   if ((order>0)&&((*ts).order==0))
   {
      vls=vertices(th,order);
      set_brs(th,order);
      first_trees(th,order);
      ntree=0;

      while (next_trees(ntree,th,order)!=0)
         ntree+=1;

      reset_nat(th,order);
   }
   else
      error(order>(*ts).order,1,"trees [tree_gen.c]",
            "Attempt to extend the tree list");

   if (order>0)
      return nall_dgm(th,order);
   else
      return 0;
}


int nbr_max(theory_t th,int order)
{
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"nbr_max [tree_gen.c]",
         "Arguments are out of range");

   ts=trs+(int)(th);
   error(order>(*ts).order,1,"nbr_max [tree_gen.c]",
         "Requested diagrams are not included in current list");

   return (*ts).nat[order];
}


int no_of_trees(theory_t th,int order,int nbr)
{
   dlist_t *dl;
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"no_of_trees [tree_gen.c]",
         "Arguments are out of range");

   if (order==0)
      return 0;

   ts=trs+(int)(th);
   error(order>(*ts).order,1,"no_of_trees [tree_gen.c]",
         "Requested diagrams are not included in current list");

   error((nbr<1)||(nbr>(*ts).nat[order]),1,"no_of_trees [tree_gen.c]",
         "Parameter nbr is out of range");

   dl=(*ts).dl[order][nbr-1];

   return (*dl).n;
}


dlist_t *select_trees(theory_t th,int order,int nbr)
{
   dlist_t *dl;
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"select_trees [tree_gen.c]",
         "Arguments are out of range");

   if (order==0)
      return NULL;

   ts=trs+(int)(th);
   error(order>(*ts).order,1,"select_trees [tree_gen.c]",
         "Requested diagrams are not included in current list");

   error((nbr<1)||(nbr>(*ts).nat[order]),1,"select_trees [tree_gen.c]",
         "Parameter nbr is out of range");

   dl=(*ts).dl[order][nbr-1];

   if ((*dl).n==0)
      return NULL;
   else
      return dl;
}


static void nena(diagram_t *dm,int *ne,int *na)
{
   int nv,n,i,j;
   int (*v)[2];
   vertex_t *vx;

   nv=(*dm).nv;
   v=(*dm).v;
   (*ne)=0;
   (*na)=-2*(*dm).nl;

   for (i=0;i<nv;i++)
   {
      vx=vls.v[v[0][0]]+v[0][1];
      n=0;

      for (j=0;j<(int)(LTYPES);j++)
         n+=(*vx).nl[j];

      if (n>(*ne))
         (*ne)=n;

      (*na)+=n;
      v+=1;
   }
}


void tree_lines(theory_t th,int order,int *nlall,int *nevtx,int *neall)
{
   int n,i,j,k;
   int nl,ne,na,*nat;
   diagram_t **dma,*dm;
   dlist_t *dl;
   tsets_t *ts;

   error((th==THEORIES)||(order<0),1,"tree_lines [tree_gen.c]",
         "Arguments are out of range");

   ts=trs+(int)(th);
   error(order>(*ts).order,1,"tree_lines [tree_gen.c]",
         "Requested diagrams are not included in current list");

   nat=(*ts).nat;
   vls=vertices(th,order);
   (*nlall)=0;
   (*nevtx)=0;
   (*neall)=0;

   for (i=1;i<=order;i++)
   {
      for (j=0;j<nat[i];j++)
      {
         dl=(*ts).dl[i][j];
         n=(*dl).n;
         dma=(*dl).dma;

         for (k=0;k<n;k++)
         {
            dm=(*dma);
            nl=(*dm).nl;
            nena(dm,&ne,&na);

            if (nl>(*nlall))
               (*nlall)=nl;
            if (ne>(*nevtx))
               (*nevtx)=ne;
            if (na>(*neall))
               (*neall)=na;

            dma+=1;
         }
      }
   }
}
